/*
 * Copyright (C) 2002-2007 The Warp Rogue Team
 * Part of the Warp Rogue Project
 *
 * This software is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License.
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY.
 *
 * See the license.txt file for more details.
 */

/*
 * Module Name: ProgramManager
 * Description: Program init/shutdown/abort functions
 */

#include <stdarg.h>
#include <time.h>

#define Uses_Ui
#define Uses_Log
#define Uses_Options
#define Uses_Npc
#define Uses_Career
#define Uses_Object
#define Uses_Macro
#define Uses_HallOfFame
#define Uses_Dice
#define Uses_Area
#define Uses_Pathfinder
#define Uses_Terrain
#define Uses_Party
#define Uses_Util
#define Uses_DataFile
#define Uses_WorldGenerator
#define Uses_RandomName
#include "mheader.h"

#include "prgman.h"

#include "dungeon.h"
#include "rdb.h"
#include "rng.h"


static void	program_init(void);
static void 	program_clean_up(void);



/*
 * initializes the game
 */
void game_init(void)
{

	program_init();
}



/*
 * game shutdown
 */
void game_shutdown(void)
{

	program_clean_up();

	exit(EXIT_SUCCESS);
}



/*
 * aborts the game in case of an internal program error
 */
void die(const char *fmt, ...)
{
	va_list vl;
	static char buffer[STRING_BUFFER_SIZE];

	va_start(vl, fmt);
	vsprintf(buffer, fmt, vl);
	va_end(vl);

	log_output(buffer);

	exit(EXIT_FAILURE);
}



/*
 * aborts the game in case of an internal program error
 * (no data path version)
 */
void die_no_data_path(const char *fmt, ...)
{
	va_list vl;
	static char buffer[STRING_BUFFER_SIZE];

	va_start(vl, fmt);
	vsprintf(buffer, fmt, vl);
	va_end(vl);

	fprintf(stderr, buffer);

	exit(EXIT_FAILURE);
}



/*
 * initializes the program
 */
static void program_init(void)
{

	log_clean_up();

	tt_seed(time(NULL));

	cpn_set_rng(random_int);

	rdb_set_error_handler(die);

	world_generator_init();

	load_data_files();

	ui_init();

	party_init();

	pathfinder_init();

	hall_of_fame_init();

	macro_init();
	
	random_name_init();
}



/*
 * does the clean up work necessary before shutdown
 */
static void program_clean_up(void)
{

	random_name_clean_up();

	hall_of_fame_clean_up();

	pathfinder_clean_up();

	world_generator_clean_up();

	area_clean_up();

	terrain_clean_up();

	npc_clean_up();

	party_clean_up();

	career_clean_up();

	object_clean_up();

	ui_clean_up();

}




