/*
 * Copyright (C) 2002-2007 The Warp Rogue Team
 * Part of the Warp Rogue Project
 *
 * This software is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License.
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY.
 *
 * See the license.txt file for more details.
 */

/*
 * Module Name: Sector
 * Description: -
 */

#define Uses_Area
#define Uses_Character
#define Uses_Object
#define Uses_Terrain
#define Uses_Dice
#define Uses_Death
#define Uses_Perception
#include "mheader.h"

#include "sector.h"



#define MAX_SECTOR_CLASSES	5



static bool	sc_free(const AREA_POINT *);

static bool	sc_free_object_location(const AREA_POINT *);
static bool	sc_free_object_location_antiblock(const AREA_POINT *);

static bool	sc_move_target(const AREA_POINT *);
static bool	sc_move_target_safe(const AREA_POINT *);


static bool	antiblock_check_passed(const AREA_POINT *);



/*
 * sector template (default values)
 */
static const SECTOR SectorTemplate = {

	/* CHARACTER */ NULL,
	/* OBJECT */ NULL,
	/* TERRAIN */ NULL

};



/*
 * sector class functions
 */
static bool (* const SectorClass[MAX_SECTOR_CLASSES])(const AREA_POINT *) = {

	sc_free,

	sc_free_object_location,
	sc_free_object_location_antiblock,

	sc_move_target,
	sc_move_target_safe

};



/*
 * initializes a sector
 */
void sector_init(SECTOR *sector)
{

	*sector = SectorTemplate;
}



/*
 * returns a string that describes a sector (used by the Look commmand)
 */
char * sector_description(char *description_string, const SECTOR *sector)
{
	const CHARACTER *character;
	const OBJECT *object;
	const TERRAIN *terrain;

	character = sector->character;
	object = sector->object;
	terrain = sector->terrain;

	if (character != NULL && (!character->is_hostile ||
		!character_unnoticed(character))) {

		strcpy(description_string, character->name);

		sprintf(description_string, "%s (%d/%d)",
			character->name,
			character->injury,
			injury_max(character)
		);

#if defined(FOR_THE_BLIND)
		if (!los(&character->location,
			&(player_controlled_character()->location))) {

			strcat(description_string, " (No LOS)");
		}
#endif

	} else if (object != NULL) {
		const OBJECT_DATA *object_data;

		object_data = object_static_data(object);

		if (object->charge != CHARGE_NIL) {

			sprintf(description_string, "%s [%d]",
				object_data->name,
				object->charge
			);

		} else {

			sprintf(description_string, "%s",
				object_data->name
			);
		}

	} else if (terrain != NULL) {
		const TERRAIN_DATA *terrain_data;

		terrain_data = terrain_static_data(terrain);

		strcpy(description_string, terrain_data->name);

		if (terrain->gore_level > GORE_LEVEL_ZERO) {

			strcat(description_string, " (");
			strcat(description_string,
				gore_level_name(terrain->gore_level)
			);
			strcat(description_string, ")");
		}
	}

	return description_string;
}



/*
 * returns true if the passed sector belongs to the passed class
 */
bool sector_is_class(const AREA_POINT *sector_location,
	SECTOR_CLASS sector_class
)
{

	return (*SectorClass[sector_class])(sector_location);
}



/*
 * sector class 'free'
 * i.e. a sector with plain terrain and without an object or a character
 */
static bool sc_free(const AREA_POINT *sector_location)
{
	const SECTOR *sector;

	sector = sector_at(sector_location);

	if (sector->character != NULL) {

		return false;
	}

	if (sector->object != NULL) {

		return false;
	}

	if (terrain_has_attribute(sector->terrain, TA_IMPASSABLE)) {

		return false;
	}

	if (terrain_has_attribute(sector->terrain, TA_DANGEROUS)) {

		return false;
	}

	return true;
}



/*
 * sector class 'free object location'
 * i.e. one can place an object there
 */
static bool sc_free_object_location(const AREA_POINT *sector_location)
{
	const SECTOR *sector;

	sector = sector_at(sector_location);

	if (sector->object != NULL) {

		return false;
	}

	if (terrain_has_attribute(sector->terrain, TA_IMPASSABLE)) {

		return false;
	}

	if (terrain_has_attribute(sector->terrain, TA_DANGEROUS)) {

		return false;
	}

	return true;
}



/*
 * sector class 'free object location antiblock'
 * i.e. one can place an impassable object there without blocking
 *      access to other sectors
 */
static bool sc_free_object_location_antiblock(const AREA_POINT *sector_location)
{

	if (sc_free_object_location(sector_location) &&
		antiblock_check_passed(sector_location)) {

		return true;
	}

	return false;
}



/*
 * sector class 'move target'
 * i.e. a character can move there
 */
static bool sc_move_target(const AREA_POINT *sector_location)
{
	const SECTOR *sector;

	sector = sector_at(sector_location);

	if (sector->character != NULL) {

		return false;
	}

	if (sector->object != NULL &&
		object_has_attribute(sector->object, OA_IMPASSABLE)) {

		return false;
	}

	if (terrain_has_attribute(sector->terrain, TA_IMPASSABLE)) {

		return false;
	}

	return true;
}



/*
 * sector class 'move target safe'
 * i.e. a character can safely move there
 */
static bool sc_move_target_safe(const AREA_POINT *sector_location)
{
	const SECTOR *sector;

	sector = sector_at(sector_location);

	if (sector->character != NULL) {

		return false;
	}

	if (sector->object != NULL &&
		object_has_attribute(sector->object, OA_IMPASSABLE)) {

		return false;
	}

	if (terrain_has_attribute(sector->terrain, TA_IMPASSABLE)) {

		return false;
	}

	if (terrain_has_attribute(sector->terrain, TA_DANGEROUS)) {

		return false;
	}

	return true;
}



/*
 * antiblock check
 */
static bool antiblock_check_passed(const AREA_POINT *sector_location)
{
	int y_mod, x_mod;

	for (y_mod = -1; y_mod <= 1; y_mod++) {
	for (x_mod = -1; x_mod <= 1; x_mod++) {
		AREA_POINT p;

		if (x_mod == 0 && y_mod == 0) {

			continue;
		}

		p.y = sector_location->y + y_mod;
		p.x = sector_location->x + x_mod;

		if (out_of_area_bounds(&p)) {

			return false;
		}

		if (!sc_move_target(&p)) {

			return false;
		}
	}
	}

	return true;
}

