/*
 * Copyright (C) 2002-2007 The Warp Rogue Team
 * Part of the Warp Rogue Project
 *
 * This software is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License.
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY.
 *
 * See the license.txt file for more details.
 */

/*
 * Module Name: WorldGenerator
 * Description: -
 */

#define Uses_Ui
#define Uses_World
#define Uses_Area
#define Uses_Sector
#define Uses_Terrain
#define Uses_Object
#define Uses_Character
#define Uses_CharacterGeneration
#define Uses_Util
#define Uses_ProgramManager
#define Uses_LoadSave
#define Uses_Npc
#define Uses_Party
#include "mheader.h"

#include "worldgen.h"

#include "dungeon.h"


#define MIN_SLUDGE_POOLS		2
#define MAX_SLUDGE_POOLS		6
#define MIN_SLUDGE_POOL_EXTENSION	0
#define MAX_SLUDGE_POOL_EXTENSION	5

#define DUNE_FREQUENCY_MIN		1
#define DUNE_FREQUENCY_MAX		100
#define DUNE_FREQUENCY			100

#define HIVE_CITY_NAME			"Hive Cruor"
#define HIVE_CITY_Y_OFFSET		1
#define HIVE_CITY_X_OFFSET		8
#define HIVE_CITY_GATE_HEIGHT		5
#define HIVE_CITY_GATE_SIDE_EX		2
#define HIVE_CITY_N_NPCS_PER_LEVEL	12
#define HIVE_CITY_N_LEVELS		(WORLD_MAX_Z + 1)

#define MIN_DISTANCE_BETWEEN_ELEVATORS	40


static void		generate_world(void);

static void		hive_city_generate(void);

static void		hive_city_generate_level(void);
static void		hive_city_generate_sector(const AREA_POINT *, CPN_TILE);

static void		hive_city_build_gate(void);

static bool		out_of_hive_bounds(const AREA_POINT *);

static void		empty_to_ash_wastes(void);
static void		empty_to_smog(void);

static void		add_elevators(void);

static void		add_chemical_sludge(void);
static void		add_chemical_sludge_pool(void);

static void		spawn_uniques(void);



/*
 * the spawn lists
 */
static LIST **			SpawnList = NULL;
static LIST **			UniquesSpawnList = NULL;


/*
 * the player character's spawn position
 */
static WORLD_POINT		PcSpawnPosition;


/*
 * the area section occupied by the hive city
 * (dungeon generator dependent)
 */
static const AREA_SECTION	HiveCityArea = {
	
	HIVE_CITY_Y_OFFSET, 61,
	HIVE_CITY_X_OFFSET, 118

};




/*
 * world generator init
 */
void world_generator_init(void)
{
	int i;

	SpawnList = checked_malloc(
		HIVE_CITY_N_LEVELS * sizeof *SpawnList
	);

	UniquesSpawnList = checked_malloc(
		HIVE_CITY_N_LEVELS * sizeof *SpawnList
	);

	for (i = 0; i < HIVE_CITY_N_LEVELS; i++) {

		SpawnList[i] = NULL;
		UniquesSpawnList[i] = NULL;
	}
}



/*
 * world generator clean up
 */
void world_generator_clean_up(void)
{

	if (SpawnList != NULL) {
		int i;

		for (i = 0; i < HIVE_CITY_N_LEVELS; i++) {

			if (SpawnList[i] != NULL) {

				list_destroy_with(SpawnList[i], free);
			}
		}

		free(SpawnList);
	}

	if (UniquesSpawnList != NULL) {
		int i;

		for (i = 0; i < HIVE_CITY_N_LEVELS; i++) {

			if (UniquesSpawnList[i] != NULL) {

				list_destroy_with(UniquesSpawnList[i], free);
			}
		}

		free(UniquesSpawnList);
	}
}


#if defined(FOR_THE_BLIND)

/*
 * the world generation screen
 */
void world_generation(void)
{
	command_bar_clear();
	render_world_generation_screen(false);
	update_screen();

	generate_world();
	
	command_bar_set(1, CM_OK);
	render_world_generation_screen(true);
	update_screen();
	
	command_bar_get_command();
}

#else 

/*
 * the world generation screen
 */
void world_generation(void)
{
	command_bar_clear();

	render_world_generation_screen(false);
	update_screen();

	generate_world();
}

#endif



/*
 * (re)spawns the NPCs
 */
void spawn_npcs(void)
{
	int m, i;
	LIST *spawn_list;

	spawn_list = SpawnList[active_area()->coord.z];
	if (spawn_list == NULL) return;

	m = spawn_list->n_nodes - 1;
	if (m < 0) return;

	for (i = area_n_characters() - party_n_members();
		i < HIVE_CITY_N_NPCS_PER_LEVEL;
		i++) {
		CHARACTER *npc;
		NPC_INDEX npc_index;
		AREA_POINT spawn_point;
		int r;

		r = random_int(0, m);

		npc_index = *((NPC_INDEX *)list_data_at(spawn_list, r));

		if (npc_index == NPC_RANDOM_HUMAN) {
			
			npc = create_random_human();

		} else {
			
			npc = npc_create_i(npc_index);
		}

		random_sector(&spawn_point, SC_MOVE_TARGET_SAFE);

		place_character(npc, &spawn_point);
	}
}




/*
 * adds a NPC to a spawn list
 */
void spawn_list_add(WORLD_COORD hive_level, const char *npc_name)
{
	NPC_INDEX *npc_index;

	npc_index = checked_malloc(sizeof *npc_index);
	
	if (strings_equal(npc_name, "RANDOM_HUMAN")) {
		
		*npc_index = NPC_RANDOM_HUMAN;

	} else {
		
		*npc_index = name_to_npc_index(npc_name);
	}

	if (npc_is_unique(*npc_index)) {

		if (UniquesSpawnList[hive_level] == NULL) {

			UniquesSpawnList[hive_level] = list_new();
		}

		list_add(UniquesSpawnList[hive_level], npc_index);

		return;
	}

	if (SpawnList[hive_level] == NULL) {

		SpawnList[hive_level] = list_new();
	}

	list_add(SpawnList[hive_level], npc_index);
}



/*
 * sets the PC's spawn position
 */
void set_pc_spawn_position(WORLD_COORD hive_level)
{

	PcSpawnPosition.z = hive_level;
	PcSpawnPosition.y = 0;
	PcSpawnPosition.x = 0;
}



/*
 * returns the PC's spawn position
 */
const WORLD_POINT * pc_spawn_position(void)
{

	return &PcSpawnPosition;
}



/*
 * generates the world
 */
static void generate_world(void)
{

	hive_city_generate();
}



/*
 * generates the hive city
 */
static void hive_city_generate(void)
{
	WORLD_POINT p;

	for (p.z = 0, p.y = 0, p.x = 0;
		p.z < HIVE_CITY_N_LEVELS;
		p.z++) {
		AREA *area;

		area = area_new();
		area->coord = p;
		set_active_area(area);

		hive_city_generate_level();

		area_write(NULL);

		area_destroy();
	}
}



/*
 * generates a hive city level
 */
static void hive_city_generate_level(void)
{
	AREA *area;
	CPN_DUNGEON *dungeon;
	int y, x;

	area = active_area();

	strcpy(area->name, HIVE_CITY_NAME);

	dungeon = cpn_build_dungeon(5, 5, NULL);
	if (dungeon == NULL) die(cpn_get_error());

	for (y = 0; y < dungeon->height; y++) {
	for (x = 0; x < dungeon->width; x++) {
		AREA_POINT p;

		p.y = y + HIVE_CITY_Y_OFFSET;
		p.x = x + HIVE_CITY_X_OFFSET;

		hive_city_generate_sector(&p, dungeon->map[y][x]);
	}
	}

	cpn_free_dungeon(dungeon);

	if (area->coord.z > 0) {

		empty_to_smog();

	} else {

		hive_city_build_gate();

		empty_to_ash_wastes();
	}

	add_elevators();

	add_chemical_sludge();

	spawn_uniques();

	spawn_npcs();
}



/*
 * generates a hive city sector
 */
static void hive_city_generate_sector(const AREA_POINT *p, CPN_TILE dungeon_tile)
{
	TERRAIN *terrain;
	OBJECT *object;

	object = NULL;

	terrain = terrain_create("Metal floor");
	place_terrain(terrain, p);

	if (dungeon_tile == CPN_FLOOR) return;

	if (dungeon_tile == CPN_WALL) {

		object = object_create("Metal wall");

	} else if (dungeon_tile == CPN_DESTRUCTABLE_OBSTACLE) {

		object = object_create("Barbed wire");

	} else if (dungeon_tile == CPN_DESTRUCTABLE_OBSTACLE_2) {

		object = object_create("Barricade");

	} else {

		die("*** CORE ERROR *** dungeon conversion failed");

	}

	if (object != NULL) {
		place_object(object, p);
	}

}



/*
 * builds the hive city gate
 */
static void hive_city_build_gate(void)
{
	AREA_POINT c;
	int y_mod;
	bool connection_reached;

	c.y = HiveCityArea.bottom;
	c.x = HiveCityArea.left +
		((HiveCityArea.right - HIVE_CITY_X_OFFSET) / 2);

	y_mod = 0;
	connection_reached = false;

	do {
		int x_mod;

		for (x_mod = -(HIVE_CITY_GATE_SIDE_EX);
			x_mod <= HIVE_CITY_GATE_SIDE_EX;
			x_mod++) {
			OBJECT *object;
			AREA_POINT p;

			p.y = c.y + y_mod;
			p.x = c.x + x_mod;

			object = object_at(&p);

			if (object == NULL) {

				connection_reached = true;

			} else {

				object_destroy(object);
				
				sector_at(&p)->object = NULL;
			}
		}

		--y_mod;

	} while (!connection_reached ||
		y_mod > -(HIVE_CITY_GATE_HEIGHT)
	);
}



/*
 * returns true if the passed point is out of bounds
 */
static bool out_of_hive_bounds(const AREA_POINT *p)
{

	if (p->y < HiveCityArea.top ||
		p->y > HiveCityArea.bottom ||
		p->x < HiveCityArea.left ||
		p->x > HiveCityArea.right) {

		return true;
	}

	return false;
}



/*
 * turns all empty sectors into ash wastes
 */
static void empty_to_ash_wastes(void)
{
	AREA_POINT p;

	for (p.y = 0; p.y < AREA_HEIGHT; p.y++) {
	for (p.x = 0; p.x < AREA_WIDTH; p.x++) {
		TERRAIN *terrain;

		if (sector_at(&p)->terrain != NULL) continue;

		if (random_int(DUNE_FREQUENCY_MIN, DUNE_FREQUENCY_MAX) ==
			DUNE_FREQUENCY) {

			terrain = terrain_create("Ash dune");
		} else {
			terrain = terrain_create("Ash wastes");
		}

		place_terrain(terrain, &p);
	}
	}

}



/*
 * turns all empty sectors into smog
 */
static void empty_to_smog(void)
{
	AREA_POINT p;

	for (p.y = 0; p.y < AREA_HEIGHT; p.y++) {
	for (p.x = 0; p.x < AREA_WIDTH; p.x++) {
		TERRAIN *terrain;

		if (sector_at(&p)->terrain != NULL) continue;

		terrain = terrain_create("Smog");
		place_terrain(terrain, &p);
	}
	}

}



/*
 * adds elevators
 */
static void add_elevators(void)
{
	AREA_POINT location1, location2;
	OBJECT *object;

	do {

		random_sector(&location1, SC_FREE_OBJECT_LOCATION_ANTIBLOCK);

	} while (out_of_hive_bounds(&location1));

	object = object_create("Elevator");
	place_object(object, &location1);

	do {

		random_sector(&location2,
			SC_FREE_OBJECT_LOCATION_ANTIBLOCK
		);

	} while (out_of_hive_bounds(&location2) ||
		area_distance(&location1, &location2) <
			MIN_DISTANCE_BETWEEN_ELEVATORS
	);

	object = object_create("Elevator");
	place_object(object, &location2);
}



/*
 * adds chemical sludge
 */
static void add_chemical_sludge(void)
{
	int i;

	for (i = random_int(MIN_SLUDGE_POOLS, MAX_SLUDGE_POOLS);
		i > 0;
		--i) {

		add_chemical_sludge_pool();
	}

}



/*
 * adds a chemical sludge pool
 */
static void add_chemical_sludge_pool(void)
{
	TERRAIN *terrain;
	AREA_POINT pool_centre;
	DIRECTION direction[MAX_DIRECTIONS];
	int i, j;

	random_sector(&pool_centre, SC_FREE);

	terrain_destroy(terrain_at(&pool_centre));
	sector_at(&pool_centre)->terrain = NULL;

	terrain = terrain_create("Chemical sludge");
	place_terrain(terrain, &pool_centre);

	randomized_directions(direction);

	for (i = random_int(MIN_SLUDGE_POOL_EXTENSION,
		MAX_SLUDGE_POOL_EXTENSION), j = 0;
		i > 0;
		--i, ++j) {
		AREA_POINT pool_extension;

		pool_extension = pool_centre;
		coordinate_step(&pool_extension, direction[j]);

		if (out_of_area_bounds(&pool_extension)) continue;

		if (sector_is_class(&pool_extension, SC_FREE)) {

			terrain_destroy(terrain_at(&pool_extension));
			sector_at(&pool_extension)->terrain = NULL;

			terrain = terrain_create("Chemical sludge");
			place_terrain(terrain, &pool_extension);
		}

	}

}



/*
 * spawns the uniques
 */
static void spawn_uniques(void)
{
	LIST *spawn_list;
	LIST_NODE *node;

	if (world_points_equal(&active_area()->coord,
		&PcSpawnPosition)) {
		AREA_POINT spawn_point;

		area_elevator_location(&spawn_point);
		adjacent_sector(&spawn_point, SC_MOVE_TARGET_SAFE);
		place_character(player_character(), &spawn_point);
	}

	spawn_list = UniquesSpawnList[active_area()->coord.z];

	if (spawn_list == NULL) {

		return;
	}

	for (node = spawn_list->head; node != NULL; node = node->next) {
		CHARACTER *npc;
		NPC_INDEX npc_index;
		AREA_POINT spawn_point;

		npc_index = *((NPC_INDEX *)node->data);

		npc = npc_create_i(npc_index);

		random_sector(&spawn_point, SC_MOVE_TARGET_SAFE);

		place_character(npc, &spawn_point);
	}
}


